import ast
from typing import List, Tuple, Union

from ..helpers import MathEvaluator


class Comparator:
    """
    Comparator class for comparing extracted answers with correct answers.

    Attributes:
        math_evaluator (MathEvaluator): An instance of MathEvaluator class.
    """

    def __init__(self):
        self.math_evaluator = MathEvaluator()

    def compare(
        self, ext_cor_pair: Tuple[str, Union[str, list], str, str]
    ) -> List[Union[str, int]]:
        """Compare the extracted answer with the correct answer. Return a list of the comparison result.

        Args:
            ext_cor_pair (Tuple[str, Union[str, list], str, str]): A tuple of the extracted answer, correct answer, and the key answer type.

        Returns:
            List[Union[str, int]]: A list of the comparison result.
        """
        right_flag = 0
        key_answer_type, standard_answer_range, extracted, correct = ext_cor_pair
        if key_answer_type == "math":
            if self.math_evaluator.is_equiv(extracted, correct) == True:
                right_flag = 1
        else:
            if extracted.strip().rstrip(".").lower() == correct.strip().rstrip(
                    ".").lower():
                right_flag = 1

            elif key_answer_type == "alphabet_option":
                if type(standard_answer_range) == str:
                    standard_answer_range_list = ast.literal_eval(
                        standard_answer_range)
                else:  
                    standard_answer_range_list = standard_answer_range
                for option in standard_answer_range_list:
                    if option[0] == correct and \
                            extracted.strip().rstrip(".").lower() == option[1].strip().rstrip(".").lower():
                        right_flag = 1
                        break

        return [*ext_cor_pair, right_flag]



import re


def normalize_final_answer(final_answer: str) -> str:
    """Normalize a final answer to a quantitative reasoning question.
    
    Args:
        final_answer (str): The final answer to normalize.
    
    Returns:
        str: The normalized final answer.
    """
    # final_answer = final_answer.split('=')[-1]
    SUBSTITUTIONS = [('an ', ''), ('a ', ''), ('.$', '$'), ('\\$', ''),
                     (r'\ ', ''), (' ', ''), ('mbox', 'text'),
                     (',\\text{and}', ','), ('\\text{and}', ','),
                     ('\\text{m}', '\\text{}'), ('\\le', '<')]
    REMOVED_EXPRESSIONS = [
        'square', 'ways', 'integers', 'dollars', 'mph', 'inches', 'ft', 'hours',
        'km', 'units', '\\ldots', 'sue', 'points', 'feet', 'minutes', 'digits',
        'cents', 'degrees', 'cm', 'gm', 'pounds', 'meters', 'meals', 'edges',
        'students', 'childrentickets', 'multiples', '\\text{s}', '\\text{.}',
        '\\text{\ns}', '\\text{}^2', '\\text{}^3', '\\text{\n}', '\\text{}',
        r'\mathrm{th}', r'^\circ', r'^{\circ}', r'\;', r',\!', '{,}', '"',
        '\\dots', '\n', '\r', '\f'
    ]
    for before, after in SUBSTITUTIONS:
        final_answer = final_answer.replace(before, after)
    for expr in REMOVED_EXPRESSIONS:
        final_answer = final_answer.replace(expr, '')

    # Extract answer that is in LaTeX math, is bold,
    # is surrounded by a box, etc.
    final_answer = re.sub(r'(\\text\{)\((.*?)\)(\})', '\\2', final_answer)
    final_answer = re.sub(r'(\\text\{)(.*?)(\})', '\\2', final_answer)
    final_answer = re.sub(r'(\\textbf\{)(.*?)(\})', '\\2', final_answer)
    final_answer = re.sub(r'(\\overline\{)(.*?)(\})', '\\2', final_answer)
    final_answer = re.sub(r'(\\boxed\{)(.*)(\})', '\\2', final_answer)
    assert '\n' not in final_answer
    assert '\r' not in final_answer
    assert '\f' not in final_answer

    if len(re.findall(r'finalansweris(.*)', final_answer)) > 0:
        final_answer = re.findall(r'finalansweris(.*)', final_answer)[-1]

    if len(re.findall(r'answer?is:?(.*)', final_answer)) > 0:
        final_answer = re.findall(r'answer?is:?(.*)', final_answer)[-1]

    if len(re.findall(r'oxed\{(.*?)\}', final_answer)) > 0:
        final_answer = re.findall(r'oxed\{(.*?)\}', final_answer)[-1]

    if len(re.findall(r'\$(.*?)\$', final_answer)) > 0:
        final_answer = re.findall(r'\$(.*?)\$', final_answer)[-1]
    final_answer = final_answer.strip()
    if 'rac' in final_answer and '\\frac' not in final_answer:
        final_answer = final_answer.replace('rac', '\\frac')

    # Normalize shorthand TeX:
    # \fracab -> \frac{a}{b}
    # \frac{abc}{bef} -> \frac{abc}{bef}
    # \fracabc -> \frac{a}{b}c
    # \sqrta -> \sqrt{a}
    # \sqrtab -> sqrt{a}b
    final_answer = re.sub(r'(frac)([^{])(.)', 'frac{\\2}{\\3}', final_answer)
    final_answer = re.sub(r'(sqrt)([^{])', 'sqrt{\\2}', final_answer)
    final_answer = final_answer.replace('$', '')

    # Normalize 100,000 -> 100000
    if final_answer.replace(',', '').isdigit():
        final_answer = final_answer.replace(',', '')

    return final_answer


class MathEvaluator():

    def __init__(self, version='v2'):
        assert version in ['v1', 'v2']
        self.version = version

    def _fix_fracs(self, string):
        """Fixes fractions in the string.
        
        Args:
            string (str): The string to fix.
            
        Returns:
            str: The fixed string.
        """
        substrs = string.split('\\frac')
        new_str = substrs[0]
        if len(substrs) > 1:
            substrs = substrs[1:]
            for substr in substrs:
                new_str += '\\frac'
                if len(substr) > 0 and substr[0] == '{':
                    new_str += substr
                else:
                    try:
                        assert len(substr) >= 2
                    except AssertionError:
                        return string
                    a = substr[0]
                    b = substr[1]
                    if b != '{':
                        if len(substr) > 2:
                            post_substr = substr[2:]
                            new_str += '{' + a + '}{' + b + '}' + post_substr
                        else:
                            new_str += '{' + a + '}{' + b + '}'
                    else:
                        if len(substr) > 2:
                            post_substr = substr[2:]
                            new_str += '{' + a + '}' + b + post_substr
                        else:
                            new_str += '{' + a + '}' + b
        string = new_str
        return string

    def _fix_a_slash_b(self, string):
        """Fixes a/b to \frac{a}{b}.
        
        Args:
            string (str): The string to fix.
            
        Returns:
            str: The fixed string.
        """
        if len(string.split('/')) != 2:
            return string
        a = string.split('/')[0]
        b = string.split('/')[1]
        try:
            a = int(a)
            b = int(b)
            assert string == '{}/{}'.format(a, b)
            new_string = '\\frac{' + str(a) + '}{' + str(b) + '}'
            return new_string
        except AssertionError:
            return string

    def _remove_right_units(self, string):
        # "\\text{ " only ever occurs (at least in the val set) when describing
        # units
        if '\\text{ ' in string:
            splits = string.split('\\text{ ')
            assert len(splits) == 2
            return splits[0]
        else:
            return string

    def _fix_sqrt(self, string):
        if '\\sqrt' not in string:
            return string
        splits = string.split('\\sqrt')
        new_string = splits[0]
        for split in splits[1:]:
            if split[0] != '{':
                a = split[0]
                new_substr = '\\sqrt{' + a + '}' + split[1:]
            else:
                new_substr = '\\sqrt' + split
            new_string += new_substr
        return new_string

    def _fix_sqrt_v2(self, string):
        _string = re.sub(r'\\sqrt(\w+)', r'\\sqrt{\1}', string)
        return _string

    def _strip_string(self, string):
        """Strip a string of unnecessary characters.
        
        Args:
            string (str): The string to strip.
            
        Returns:
            str: The stripped string.
        """
        # linebreaks
        string = string.replace('\n', '')

        # remove inverse spaces
        string = string.replace('\\!', '')

        # replace \\ with \
        string = string.replace('\\\\', '\\')

        # replace tfrac and dfrac with frac
        string = string.replace('tfrac', 'frac')
        string = string.replace('dfrac', 'frac')

        # remove \left and \right
        string = string.replace('\\left', '')
        string = string.replace('\\right', '')

        # Remove circ (degrees)
        string = string.replace('^{\\circ}', '')
        string = string.replace('^\\circ', '')

        # remove dollar signs
        string = string.replace('\\$', '')

        # remove units (on the right)
        string = self._remove_right_units(string)

        # remove percentage
        string = string.replace('\\%', '')
        string = string.replace('\%', '')  # noqa: W605

        # " 0." equivalent to " ." and "{0." equivalent to "{." Alternatively,
        # add "0" if "." is the start of the string
        string = string.replace(' .', ' 0.')
        string = string.replace('{.', '{0.')
        # if empty, return empty string
        if len(string) == 0:
            return string
        if string[0] == '.':
            string = '0' + string

        # to consider: get rid of e.g. "k = " or "q = " at beginning
        if len(string.split('=')) == 2:
            if len(string.split('=')[0]) <= 2:
                string = string.split('=')[1]

        # fix sqrt3 --> sqrt{3}
        string = self._fix_sqrt(string)

        # remove spaces
        string = string.replace(' ', '')

        # \frac1b or \frac12 --> \frac{1}{b} and \frac{1}{2}, etc. Even works
        # with \frac1{72} (but not \frac{72}1). Also does a/b --> \\frac{a}{b}
        string = self._fix_fracs(string)

        # manually change 0.5 --> \frac{1}{2}
        if string == '0.5':
            string = '\\frac{1}{2}'

        # NOTE: X/Y changed to \frac{X}{Y} in dataset, but in simple cases fix
        # in case the model output is X/Y
        string = self._fix_a_slash_b(string)

        return string

    def _strip_string_v2(self, string):
        string = str(string).strip()
        # linebreaks
        string = string.replace('\n', '')

        # right "."
        string = string.rstrip('.')

        # remove inverse spaces
        string = string.replace('\\!', '')
        string = string.replace('\\ ', '')

        # replace \\ with \
        string = string.replace('\\\\', '\\')
        string = string.replace('\\\\', '\\')

        # replace tfrac and dfrac with frac
        string = string.replace('tfrac', 'frac')
        string = string.replace('dfrac', 'frac')

        # remove \left and \right
        string = string.replace('\\left', '')
        string = string.replace('\\right', '')

        # Remove unit: miles, dollars if after is not none
        _string = re.sub(r'\\text{.*?}$', '', string).strip()
        if _string != '' and _string != string:
            string = _string

        # Remove circ (degrees)
        string = string.replace('^{\\circ}', '')
        string = string.replace('^\\circ', '')

        # remove dollar signs
        string = string.replace('\\$', '')
        string = string.replace('$', '')

        string = string.replace('\\text', '')
        string = string.replace('x\\in', '')

        # remove percentage
        string = string.replace('\\%', '')
        string = string.replace('\%', '')  # noqa: W605
        string = string.replace('%', '')

        # " 0." equivalent to " ." and "{0." equivalent to "{." Alternatively,
        # add "0" if "." is the start of the string
        string = string.replace(' .', ' 0.')
        string = string.replace('{.', '{0.')

        # cdot
        string = string.replace('\\cdot', '')

        # inf
        string = string.replace('infinity', '\\infty')
        if '\\infty' not in string:
            string = string.replace('inf', '\\infty')
        string = string.replace('+\\inity', '\\infty')

        # and
        string = string.replace('and', '')
        string = string.replace('\\mathbf', '')

        # use regex to remove \mbox{...}
        string = re.sub(r'\\mbox{.*?}', '', string)

        # quote
        string.replace("'", '')
        string.replace('"', '')

        # i, j
        if 'j' in string and 'i' not in string:
            string = string.replace('j', 'i')

        # replace a.000b where b is not number or b is end, with ab, use regex
        string = re.sub(r'(\d+)\.0+([^\d])', r'\1\2', string)
        string = re.sub(r'(\d+)\.0+$', r'\1', string)

        # if empty, return empty string
        if len(string) == 0:
            return string
        if string[0] == '.':
            string = '0' + string

        # to consider: get rid of e.g. "k = " or "q = " at beginning
        if len(string.split('=')) == 2:
            if len(string.split('=')[0]) <= 2:
                string = string.split('=')[1]

        string = self._fix_sqrt_v2(string)
        string = string.replace(' ', '')

        # \frac1b or \frac12 --> \frac{1}{b} and \frac{1}{2}, etc.
        # Even works with \frac1{72} (but not \frac{72}1).
        # Also does a/b --> \\frac{a}{b}
        string = self._fix_fracs(string)

        # NOTE: X/Y changed to \frac{X}{Y} in dataset, but in simple
        # cases fix in case the model output is X/Y
        string = self._fix_a_slash_b(string)

        return string

    def is_equiv(self, str1, str2, verbose=False):
        """Check if two strings are equivalent.
        
        Args:
            str1 (str): The first string.
            str2 (str): The second string.
            verbose (bool): If True, print the stripped strings.
        
        Returns:
            bool: True if the strings are equivalent, False otherwise.
        """
        if str1 is None and str2 is None:
            print('WARNING: Both None')
            return True
        if str1 is None or str2 is None:
            return False

        if self.version == 'v1':
            strip_string_func = self._strip_string
        elif self.version == 'v2':
            strip_string_func = self._strip_string_v2
        else:
            raise NotImplementedError

        try:
            ss1 = strip_string_func(str1)
            ss2 = strip_string_func(str2)
            if verbose:
                print(ss1, ss2)
            if ss1 == ss2:
                return True
            ss1 = normalize_final_answer(ss1)
            ss2 = normalize_final_answer(ss2)
            if ss1 == ss2:
                return True
        except Exception:
            pass

        try:
            ss1 = normalize_final_answer(str1)
            ss2 = normalize_final_answer(str2)
            if ss1 == ss2:
                return True
        except Exception:
            pass

        # try:
        #     if is_equiv(str1, str2):
        #         return True
        # except Exception:
        #     pass

        # try:
        #     ss1 = normalize_final_answer(str1)
        #     ss2 = normalize_final_answer(str2)
        #     if is_equiv(ss1, ss2):
        #         return True
        # except Exception:
        #     pass

        return str1 == str2


if __name__ == '__main__':
    math_evaluator = MathEvaluator()
    print(math_evaluator.is_equiv("\\boxed{1005}", "1005"))



import ast
import json


class DataProcessor:

    def __init__(self):
        pass

    def read_data(self, path):
        with open(path, 'r', encoding="utf-8") as f:
            data = json.load(f)

        for item in data:
            if isinstance(item["standard_answer_range"],
                          str) and item["key_answer_type"] != "math":
                try:
                    item["standard_answer_range"] = ast.literal_eval(
                        item["standard_answer_range"])
                except Exception as e:
                    print(f"Error: {e}")
                    print(
                        "Please check if you provide the correct form of the 'standard_answer_range': ",
                        item)
                    exit(0)

            item["standard_answer_range"] = str(item["standard_answer_range"])
            item["key_answer_type"] = str(item["key_answer_type"])

        return data



import argparse
import importlib.resources as pkg_resources
import json

from tqdm import tqdm

from . import examples
from .helpers import DataProcessor
from .modules import Comparator, Extractor


class Evaluator:
    """
    Evaluator class for evaluating the performance of the xFinder model.

    Args:
        model_name (str, optional): The model name to be used for inference. Defaults to "xFinder-qwen1505".
        inference_mode (str, optional): The mode of inference, either 'local' or 'api'. Defaults to "local".
        model_path_or_url (str, optional): The path or URL of the model. Defaults to "IAAR-Shanghai/xFinder-qwen1505".
        temperature (float, optional): The temperature value for sampling. Defaults to 0.7.
        max_tokens (int, optional): The maximum number of tokens to generate. Defaults to 100.
    """

    MATH_STANDARD_ANSWER_RANGE = "a(n) number / set / vector / matrix / interval / expression / function / equation / inequality"
    VALID_KEY_ANSWER_TYPES = {"math", "short_text",
                              "categorical_label", "alphabet_option"}

    def __init__(self, model_name="xFinder-qwen1505", inference_mode="local", model_path_or_url="IAAR-Shanghai/xFinder-qwen1505", temperature=0.7, max_tokens=100):
        self.extractor = Extractor(
            model_name=model_name,
            inference_mode=inference_mode,
            model_path_or_url=model_path_or_url,
            temperature=temperature,
            max_tokens=max_tokens,
        )
        self.comparator = Comparator()
        self.data_processor = DataProcessor()

    def evaluate_single_item(self, question, llm_output, answer_range, answer_type, correct_answer):
        if answer_type not in self.VALID_KEY_ANSWER_TYPES:
            raise ValueError(
                f"Invalid key_answer_type: {answer_type}. Must be one of {self.VALID_KEY_ANSWER_TYPES}")

        if answer_type == "math":
            answer_range = self.MATH_STANDARD_ANSWER_RANGE

        extracted_answer = self.extractor.generate_output(
            question, llm_output, answer_range)
        evaluation_result = self.comparator.compare(
            (answer_type, answer_range, extracted_answer, correct_answer)
        )
        return evaluation_result

    def evaluate(self, data_path):
        data = self.data_processor.read_data(data_path)
        results = [
            self.evaluate_single_item(
                item["question"],
                item["llm_output"],
                self.MATH_STANDARD_ANSWER_RANGE if item["key_answer_type"] == "math" else item["standard_answer_range"],
                item["key_answer_type"],
                item["correct_answer"]
            ) for item in tqdm(data, desc="Evaluating")
        ]
        correct_count = sum(result[-1] for result in results)
        accuracy = correct_count / max(1, len(results)) if results else 0
        return accuracy

    def evaluate_single_example(self, question, llm_output, standard_answer_range, key_answer_type, correct_answer):
        if key_answer_type not in self.VALID_KEY_ANSWER_TYPES:
            raise ValueError(
                f"Invalid key_answer_type: {key_answer_type}. Must be one of {self.VALID_KEY_ANSWER_TYPES}")

        evaluation_result = self.evaluate_single_item(
            question,
            llm_output,
            standard_answer_range,
            key_answer_type,
            correct_answer
        )
        judgement = "Correct" if evaluation_result[-1] else "Incorrect"
        return judgement

    def save_results(self, results, output_path):
        with open(output_path, 'w') as f:
            json.dump(results, f, indent=4, ensure_ascii=False)


def run_example(model_name, inference_mode, model_path_or_url, temperature, max_tokens):
    evaluator = Evaluator(
        model_name=model_name,
        inference_mode=inference_mode,
        model_path_or_url=model_path_or_url,
        temperature=temperature,
        max_tokens=max_tokens
    )

    # example for batch evaluation
    example_path = pkg_resources.path(examples, "example.json")
    evaluator.evaluate(example_path)

    # example for single example evaluation
    question = "What is the capital of France?"
    llm_output = "The capital of France is Paris."
    standard_answer_range = "[\"Paris\", \"Lyon\", \"Marseille\"]"
    key_answer_type = "short_text"
    correct_answer = "Paris"

    # evaluate single example
    result = evaluator.evaluate_single_example(
        question,
        llm_output,
        standard_answer_range,
        key_answer_type,
        correct_answer
    )
    print(f"Single example evaluation result: {result}")


def main():
    parser = argparse.ArgumentParser(
        description="Run evaluation examples for the xFinder model.")
    parser.add_argument('--run-example', action='store_true',
                        help="Run the example evaluation.")
    parser.add_argument('--model-name', type=str, default="xFinder-qwen1505",
                        help="The model name to be used for inference.")
    parser.add_argument('--inference-mode', type=str, default="local",
                        help="The mode of inference, either 'local' or 'api'.")
    parser.add_argument('--model-path-or-url', type=str,
                        default="IAAR-Shanghai/xFinder-qwen1505", help="The path or URL of the model.")
    parser.add_argument('--temperature', type=float, default=0.7,
                        help="The temperature value for sampling.")
    parser.add_argument('--max-tokens', type=int, default=100,
                        help="The maximum number of tokens to generate.")
    args = parser.parse_args()

    if args.run_example:
        run_example(
            model_name=args.model_name,
            inference_mode=args.inference_mode,
            model_path_or_url=args.model_path_or_url,
            temperature=args.temperature,
            max_tokens=args.max_tokens
        )


if __name__ == '__main__':
    main()
